
/* Chrysalide - Outil d'analyse de fichiers binaires
 * item.c - équivalent Python du fichier "analysis/db/item.c"
 *
 * Copyright (C) 2014-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "item.h"


#include <pygobject.h>


#include <i18n.h>
#include <analysis/db/item.h>


#include "../../helpers.h"



/* Indique si l'élément contient des données à oublier ou non. */
static PyObject *py_db_item_get_volatile(PyObject *, void *);

/* Définit si l'élément contient des données à oublier ou non. */
static int py_db_item_set_volatile(PyObject *, PyObject *, void *);



/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Indique si l'élément contient des données à oublier ou non.  *
*                                                                             *
*  Retour      : Etat de la sauegarde de l'élément consulté.                  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_db_item_get_volatile(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GDbItem *item;                          /* Elément à consulter         */

    item = G_DB_ITEM(pygobject_get(self));

    result = (g_db_item_is_volatile(item) ? Py_True : Py_False);
    Py_INCREF(result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                value   = valeur fournie à intégrer ou prendre en compte.    *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Définit si l'élément contient des données à oublier ou non.  *
*                                                                             *
*  Retour      : Bilan de l'opération pour Python.                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int py_db_item_set_volatile(PyObject *self, PyObject *value, void *closure)
{
    GDbItem *item;                          /* Elément à modifier          */

    if (!PyBool_Check(value))
    {
        PyErr_SetString(PyExc_TypeError, _("The attribute value must be a boolean."));
        return -1;
    }

    item = G_DB_ITEM(pygobject_get(self));
    g_db_item_set_volatile(item, (bool)(value == Py_True));

    return 0;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_db_item_type(void)
{
    static PyMethodDef py_db_item_methods[] = {
        { NULL }
    };

    static PyGetSetDef py_db_item_getseters[] = {

        {
            "volatile", py_db_item_get_volatile, py_db_item_set_volatile,
            "Define if a Database item can be forgotten.", NULL
        },
        { NULL }

    };

    static PyTypeObject py_db_item_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.analysis.db.DbItem",
        .tp_basicsize   = sizeof(PyGObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,

        .tp_doc         = "PyChrysalide item for DataBase collection",

        .tp_methods     = py_db_item_methods,
        .tp_getset      = py_db_item_getseters,

    };

    return &py_db_item_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : module = module dont la définition est à compléter.          *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide....db.items.DbItem'.   *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool register_python_db_item(PyObject *module)
{
    PyTypeObject *py_db_item_type;          /* Type Python 'DbItem'        */
    PyObject *dict;                         /* Dictionnaire du module      */

    py_db_item_type = get_python_db_item_type();

    dict = PyModule_GetDict(module);

    if (!register_class_for_pygobject(dict, G_TYPE_DB_ITEM, py_db_item_type, &PyGObject_Type))
        return false;

    return true;

}
