
/* Chrysalide - Outil d'analyse de fichiers binaires
 * symbol.c - équivalent Python du fichier "format/symbol.h"
 *
 * Copyright (C) 2015-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "symbol.h"


#include <pygobject.h>


#include <i18n.h>


#include <format/symbol.h>


#include "../helpers.h"
#include "../analysis/routine.h"
#include "../analysis/db/items/comment.h"
#include "../arch/instruction.h"
#include "../arch/vmpa.h"



/* Effectue une comparaison avec un objet Python 'BinSymbol'. */
static PyObject *py_binary_symbol_richcompare(PyObject *, PyObject *, int);

/* Définit un autre nom pour le symbole. */
static PyObject *py_binary_symbol_set_alt_label(PyObject *, PyObject *);

/* Fournit le type du symbole. */
static PyObject *py_binary_symbol_get_target_type(PyObject *, void *);

/* Fournit un étiquette pour viser un symbole. */
static PyObject *py_binary_symbol_get_label(PyObject *, void *);

/* Fournit l'emplacement où se situe un symbole. */
static PyObject *py_binary_symbol_get_range(PyObject *, void *);

/* Définit les constantes pour les symboles binaires. */
static bool py_binary_symbol_define_constants(PyTypeObject *);



/******************************************************************************
*                                                                             *
*  Paramètres  : a  = premier object Python à consulter.                      *
*                b  = second object Python à consulter.                       *
*                op = type de comparaison menée.                              *
*                                                                             *
*  Description : Effectue une comparaison avec un objet Python 'BinSymbol'.   *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_richcompare(PyObject *a, PyObject *b, int op)
{
    PyObject *result;                       /* Bilan à retourner           */
    int ret;                                /* Bilan de lecture des args.  */
    const GBinSymbol *sym_a;                /* Premier élément à traiter   */
    const GBinSymbol *sym_b;                /* Second élément à traiter    */
    int status;                             /* Résultat d'une comparaison  */

    ret = PyObject_IsInstance(b, (PyObject *)get_python_binary_symbol_type());
    if (!ret)
    {
        result = Py_NotImplemented;
        goto cmp_done;
    }

    sym_a = G_BIN_SYMBOL(pygobject_get(a));
    sym_a = G_BIN_SYMBOL(pygobject_get(b));

    status = g_binary_symbol_cmp(&sym_a, &sym_b);

    result = status_to_rich_cmp_state(status, op);

 cmp_done:

    Py_INCREF(result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : type = type de l'objet à instancier.                         *
*                args = arguments fournis à l'appel.                          *
*                kwds = arguments de type key=val fournis.                    *
*                                                                             *
*  Description : Crée un nouvel objet Python de type 'DbComment'.             *
*                                                                             *
*  Retour      : Instance Python mise en place.                               *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    PyObject *result;                       /* Bilan à retourner           */
    SymbolType stype;                       /* Type prévu pour le  symbole */
    PyObject *range_obj;                    /* Objet pour la couverture    */
    int ret;                                /* Bilan de lecture des args.  */
    mrange_t *range;                        /* Version native d'un espace  */
    GBinSymbol *symbol;                     /* Version GLib du symble      */

    ret = PyArg_ParseTuple(args, "lO", &stype, &range_obj);
    if (!ret) return NULL;

    if (stype >= STP_COUNT)
    {
        PyErr_SetString(PyExc_ValueError, _("Invalid type of message"));
        return NULL;
    }

    ret = PyObject_IsInstance(range_obj, (PyObject *)get_python_mrange_type());
    if (!ret)
    {
        PyErr_SetString(PyExc_TypeError, _("The second argument must be an instance of mrange."));
        return NULL;
    }

    range = get_internal_mrange(range_obj);

    symbol = g_binary_symbol_new(range, stype);

    result = pygobject_new(G_OBJECT(symbol));
    g_object_unref(symbol);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Définit un autre nom pour le symbole.                        *
*                                                                             *
*  Retour      : None.                                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_set_alt_label(PyObject *self, PyObject *args)
{
    const char *alt;                        /* Etiquette alternative       */
    int ret;                                /* Bilan de lecture des args.  */
    GBinSymbol *symbol;                     /* Elément à consulter         */

    ret = PyArg_ParseTuple(args, "s", &alt);
    if (!ret) return NULL;

    symbol = G_BIN_SYMBOL(pygobject_get(self));

    g_binary_symbol_set_alt_label(symbol, alt);

    Py_RETURN_NONE;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit le type du symbole.                                  *
*                                                                             *
*  Retour      : Type de symbole représenté.                                  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_get_target_type(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinSymbol *symbol;                     /* Elément à consulter         */
    SymbolType type;                        /* Type de symbole représenté  */

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    type = g_binary_symbol_get_target_type(symbol);

    result = PyLong_FromLong(type);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit un étiquette pour viser un symbole.                  *
*                                                                             *
*  Retour      : Chaîne de caractères renvoyant au symbole.                   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_get_label(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinSymbol *symbol;                     /* Elément à consulter         */
    const char *label;                      /* Désignation courante        */

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    label = g_binary_symbol_get_label(symbol);

    if (label != NULL)
        result = PyUnicode_FromString(label);

    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit l'emplacement où se situe un symbole.                *
*                                                                             *
*  Retour      : Zone mémoire couverte par le symbole.                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_symbol_get_range(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinSymbol *symbol;                     /* Elément à consulter         */
    const mrange_t *range;                  /* Couverture courante         */

    symbol = G_BIN_SYMBOL(pygobject_get(self));
    range = g_binary_symbol_get_range(symbol);

    result = build_from_internal_mrange(range);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : obj_type = type dont le dictionnaire est à compléter.        *
*                                                                             *
*  Description : Définit les constantes pour les symboles binaires.           *
*                                                                             *
*  Retour      : true en cas de succès de l'opération, false sinon.           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool py_binary_symbol_define_constants(PyTypeObject *obj_type)
{
    bool result;                            /* Bilan à retourner           */

    result = true;

    result &= PyDict_AddIntMacro(obj_type, STP_DATA);
    result &= PyDict_AddIntMacro(obj_type, STP_ROUTINE);
    result &= PyDict_AddIntMacro(obj_type, STP_CODE_LABEL);
    result &= PyDict_AddIntMacro(obj_type, STP_OBJECT);
    result &= PyDict_AddIntMacro(obj_type, STP_ENTRY_POINT);
    result &= PyDict_AddIntMacro(obj_type, STP_STRING);
    result &= PyDict_AddIntMacro(obj_type, STP_RO_STRING);
    result &= PyDict_AddIntMacro(obj_type, STP_COUNT);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_binary_symbol_type(void)
{
    static PyMethodDef py_bin_symbol_methods[] = {
        {
            "set_alt_label", py_binary_symbol_set_alt_label,
            METH_VARARGS,
            "set_alt_label($self, alt, /)\n--\n\nSet an alternative label for the symbol."
        },
        { NULL }
    };

    static PyGetSetDef py_bin_symbol_getseters[] = {
        {
            "target_type", py_binary_symbol_get_target_type, NULL,
            "Type of the current symbol.", NULL
        },
        {
            "label", py_binary_symbol_get_label, NULL,
            "Label of the symbol, provided by the internal component or by an alternative label.", NULL
        },
        {
            "range", py_binary_symbol_get_range, NULL,
            "Range covered by the symbol.", NULL
        },
        { NULL }
    };

    static PyTypeObject py_bin_symbol_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.format.BinSymbol",
        .tp_basicsize   = sizeof(PyGObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT,

        .tp_doc         = "PyChrysalide binary symbol",

        .tp_richcompare = py_binary_symbol_richcompare,

        .tp_methods     = py_bin_symbol_methods,
        .tp_getset      = py_bin_symbol_getseters,
        .tp_new         = (newfunc)py_binary_symbol_new

    };

    return &py_bin_symbol_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : module = module dont la définition est à compléter.          *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide.format.BinSymbol'.     *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool register_python_binary_symbol(PyObject *module)
{
    PyTypeObject *py_bin_symbol_type;       /* Type Python 'BinSymbol'     */
    PyObject *dict;                         /* Dictionnaire du module      */

    py_bin_symbol_type = get_python_binary_symbol_type();

    dict = PyModule_GetDict(module);

    if (!register_class_for_pygobject(dict, G_TYPE_BIN_SYMBOL, py_bin_symbol_type, &PyGObject_Type))
        return false;

    if (!py_binary_symbol_define_constants(py_bin_symbol_type))
        return false;

    return true;

}
